{{
        Karplus-Strong on the propeller
        (sounds like a plucked string)

        Jonathan "lonesock" Dummer

        NOTES:

        * use any buffer size you want
          - no power-of-2-restriction
          - the buffer uses words (2 bytes per sample)

        * fraction_bits must be [1..8]
          - more means more accurate delay-sample interpolation
          - more means a bit more computation (12-clocks per bit)

        * the period of the plucked string is specified in fixed-point
          - if 'fraction_bits' = 4, then you should use period * 2^4,
            and the code will use the period to the nearest 1/16th of
            a sample.
          - the exact period isn't precisely determined yet..I still
            have to do some math regarding the simple filtering I used.

        * you can change the sustain of the string with 'sustain'
          - values of ~ 3..16 make sense
          - implemented as scaling the feedback by (2^N - 1) / (2^N)
          - the higher this number...the longer it will ring
          - you can change this whenever you want

        * set_note uses the MIDI note convention (69 => A440)

        * I can now use NCO mode to drive the output pin, per Lawson's suggestion
          - to make the scaling work, I set FRQx to be equal to POSX / clocks_per_period
          - you change the value by setting PHSx to a value between 0 and POSX
          - if PHSx starts at 0, adding FRQx to it each clock will yield POSX at the
            end of the period, meaning that the high bit never goes to 1
          - if PHSx starts at POSX, after the 1st clock FRQx will have been added, so the pin goes high
          - if PHSx is set to POSX/2, it will spend 1/2 the period low, then 1/2 the period high

        TODO:

        * simulate multiple strings in a single cog
}}
CON
  buf_size      = 1650           ' any size you want...just need enough RAM
  fraction_bits = 8             ' must be in the range [1..8]
  use_DUTY_mode = 1             ' use DUTY mode (otherwise uses NCO)

  ' used in my control array
  ctrl_VOLUME   = 0             ' set this to [1..31] to "pluck" a string
  ctrl_DELAY    = 1             ' set this to the period * 2^fraction_bits to set the frequency
  ctrl_SUSTAIN  = 2             ' set this to [1..31] to control how long the string rings

DAT
{{
  Matched to the MIDI spec, freq = 440 * 2 ^ ( (N - 69) / 12 )
  MIDI note ratios for 0..11 (every 12 notes goes up an octave...just divide the period by 2)
  fixed point period = sample_rate << fraction_bits / frequency
  period = sample_rate << fraction_bits * 2^32 / 2^32 / frequency
  using the ** operator (x ** y) means (x * y / 2^32)
  so if ratio = 2^32 / frequency
  period = sample_rate << fraction_bits ** ratio
}}
ratio0  long $1F4FDA4C  ' 2^32 / 8.175799 Hz
ratio1  long $1D8DF51D  ' 2^32 / 8.661957 Hz
ratio2  long $1BE5501B  ' 2^32 / 9.177024 Hz
ratio3  long $1A548077  ' 2^32 / 9.722718 Hz
ratio4  long $18DA2FC0  ' 2^32 / 10.300861 Hz
ratio5  long $17751ABC  ' 2^32 / 10.913382 Hz
ratio6  long $16241056  ' 2^32 / 11.562326 Hz
ratio7  long $14E5F099  ' 2^32 / 12.249857 Hz
ratio8  long $13B9ABB9  ' 2^32 / 12.978272 Hz
ratio9  long $129E412A  ' 2^32 / 13.750000 Hz
ratio10 long $1192BEC8  ' 2^32 / 14.567618 Hz
ratio11 long $10964004  ' 2^32 / 15.433853 Hz

VAR
  long  control[3]              ' here's where I pass commands to the PASM cog (PAR)
  long  sample_rate_fixed       ' sample-rate << fraction_bits, used in pluck_note
  word  buf[buf_size]           ' using words to minimize the memory cost
  byte  cog                     ' which cog (if any) am I using?

{ simple test code...remove the leading ' to comment this block out
CON
  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000
  sr            = 48_000        ' let's run at > CD sampling rate
  sustain_value = 8
  demo_pinL     = 10
  demo_pinR     = 11
OBJ
  term          : "FullDuplexSerial"
PUB test_KS | rn, p

  ' set up my serial port, and wait for the BST terminal to go live
  term.start( 31, 30, 0, 115_200 )

  ' set up using the demo board pins for the headphones
  start( demo_pinL, demo_pinR, sr )

  waitcnt( clkfreq >> 3 + cnt )

  { let's just hear a single note
  set_sustain( 8 )
  'set_frequency( 44100 / 6 )
  set_note( 69 )
  pluck( 255 )
  repeat
  '}

  set_sustain( sustain_value )

  { let's look at the waveform!
  'set_frequency( 44100 / 6 )
  set_note( 69 )

  p := clkfreq * 3 / sr * buf_size / 4
  pluck( 255 )
  'waitcnt( p + cnt )
  waitcnt( clkfreq * 2 + cnt )
  stop
  'term.tx( 0 )
  term.str( string( 13, "period (fixed point) = " ) )
  term.dec( control[ctrl_DELAY] )
  term.str( string( 13, "index,value", 13 ) )
  repeat p from 0 to constant( buf_size - 1 )
  'repeat p from 575 to 585
    if buf[p]
      term.dec( p )
      term.tx( "," )
      term.dec( ~~buf[p] )
      term.tx( 13 )
  repeat
  '}

  { check the volume
  set_note( 69 )
  p := 0
  repeat
    ' wait a bit
    waitcnt( clkfreq / 4 + cnt )
    p := (p + 10) & 255
    pluck( p )
  '}

  ' play the scale
  rn := 0

  repeat
    ' wait a bit
    waitcnt( clkfreq / 4 + cnt )
    ' pluck a note (MIDI spec) and returns the period
    set_note( rn )
    if( rn & %11 == 0 )
      pluck( 255 )
    ' debug
    term.dec( rn )
    term.tx( 13 )
    ' next note?
    ++rn
    rn //= 128
'}

{{
  The minimum period is < 500 clocks, so keep the sample rate < clock / 500
  So, for the prop running at 80MHz, the sample rate should be less that 160 kHz [8^)
}}
PUB start( pinL, pinR, sample_rate )
  stop
  ' set up the counter and pins
  if (pinL < 0) AND (pinR < 0)
    return 0
  if (pinL > -1) AND (pinR > -1)
    ctra_val := constant( |<26 ) | (pinL << 9) | pinR
    control := |<pinL | |<pinR
  else
    if pinL > -1
      ctra_val := pinL
      control := |<pinL
    else
      ctra_val := pinR
      control := |<pinR
  ' and the counter mode?
  if use_DUTY_mode
    ctra_val |= constant( %00110 << 26 )                ' DUTY
  else
    ctra_val |= constant( %00100 << 26 )                ' NCO
  ' set up the pseudo-constants used in the PASM code
  sample_rate_fixed := sample_rate << fraction_bits
  period := (clkfreq + (sample_rate >> 1)) / sample_rate
  NCO_stride := posx / period   ' only used in NCO mode
  buf_head := @buf
  buf_tail := buf_head + constant(buf_size * 2)
  max_delay := constant( (buf_size - 2) << fraction_bits )
  ' start it up
  control[ctrl_DELAY] := constant( 1 << fraction_bits )
  control[ctrl_SUSTAIN] := 8
  cog := cognew( @ks_PASM, @control ) + 1
  return cog

PUB set_sustain( sustain )
  control[ctrl_SUSTAIN] := sustain <# 31

PUB set_frequency( Hz )
  control[ctrl_DELAY] := (sample_rate_fixed / Hz) <# max_delay

PUB set_note( MIDI_note )
  control[ctrl_DELAY] := ((sample_rate_fixed ** ratio0[MIDI_note//12]) >> (MIDI_note / 12)) <# max_delay
'{ the simple volume-only version
PUB pluck( volume )
  ' volume will be clamped to 255 max
  control := volume
'}
{ the faux-sine version
PUB pluck( volume ) : p
  ' volume goes in the low byte, and the pseudo-sine acceleration term goes in the upper 24 bits
  p := control[ctrl_DELAY] >> constant( fraction_bits + 2 )
  control := (volume <# 255) | ((constant( |<SINE_SHL >> 7 ) / p / p) << 8)
'}

PUB dump_buf( index )
  return ~~buf[index <# constant( buf_size - 1 )]

PUB stop
  if cog
    cogstop( cog~ - 1 )

DAT
ORG 0

ks_PASM       ' setup
              mov       rnd_seed, cnt
                        '
              mov       frqa, NCO_stride
              mov       ctra, ctra_val

              rdlong    dira, par
              mov       timestamp, period
              add       timestamp, cnt
              wrlong    phsb, par
              mov       ptr_write, buf_head
              mov       ptr_read, buf_head

loop          ' main loop: wait for the proper time, then after a constant delay set the output value (no rd*/wr* yet!)
              waitcnt   timestamp, period
              add       sample, middle_val

              '          which counter mde am I using?
              mov       tmp, #use_DUTY_mode wz
              ' NCO
        if_z  shl       sample, #15
        if_z  mov       phsa, sample
              ' DUTY
        if_nz shl       sample, #16
        if_nz mov       frqa, sample

              ' OK, what are my operational parameters?
              mov       tmp, par
              add       tmp, #8
              rdlong    sustain_sar, tmp
              max       sustain_sar, #31
              sub       tmp, #4
              rdlong    int_delay, tmp
              ' how much should I shift a random number so it dithers the low bit before sustain_sar?
              ' note: N (0..31) XOR %11111 = 31 - N
              mov       sustain_rnd, #%11111
              xor       sustain_rnd, sustain_sar

              ' update those settings, whether they changed or not
              call      #adjust_delay

              ' does the user want to pluck the string?
              rdlong    tmp, par wz
        if_z  jmp       #skip_init_pluck
              mov       pluck_vol, tmp
              rev       pluck_vol, #32-8
              wrlong    phsb, par
              mov       pluck_left, int_delay

skip_init_pluck

              ' perform the delay / filter / feedback pointer add & wrap
              mov       last_val, this_val
              add       ptr_read, #2
              cmpsub    ptr_read, buf_tail wc
        if_c  add       ptr_read, buf_head

              ' if not plucking, read this value and sign-extend
              rdword    this_val, ptr_read
              shl       this_val, #16
              sar       this_val, #16

              ' if plucking, generate a pseudo-random number as a white-noise input.
              '{ Ahle2's version
              add       rnd_seed, prng_add      ' $80808080
              ror       rnd_seed, #17
              '}
              { Chip's version
              min       rnd_seed, #1
              test      rnd_seed, #%10111 wc
              rcr       rnd_seed, #1
              '}
              { Philip Koopman's version: http://www.ece.cmu.edu/~koopman/lfsr/index.html
              min       rnd_seed, #1
              ror       rnd_seed, #1 wc
        if_c  xor       rnd_seed, #$57
              '}

              ' Are we plucking?  C is set if we are plucking
              cmpsub    pluck_left, #1 wc
        if_nc jmp       #:continue

              mov       tmpt, rnd_seed          ' start with a 32-bit random value
              shr       tmpt, #8                ' convert it to a 24-bit random value
              shl       tmpt, #8                ' and make the low 8 bits
              or        tmpt, pluck_vol         ' my 8-bit multiplier

              ' here is my volume control
              'shl       this_val, #16            ' start with my accumulator at the old value
              mov       this_val, #0            ' start with my accumulator at 0
              mov       tmp, #8                 ' 8 bits

:vol_loop     sar       tmpt, #1 wc
        if_c  add       this_val, tmpt
              djnz      tmp, #:vol_loop
              sar       this_val, #16

:continue

{{
              This section does a linear interpolation.
              BUT
              We need filtered samples to sound OK, so we first do a simple
              average-of-2-sample lowpass, then we interpolate between 2 of
              those filtered samples!
}}
              ' shuffle the most recent filtered value into it's correct place
              mov       last_filter, this_filter
              ' and calculate the newest filtered value
              mov       this_filter, this_val
              add       this_filter, last_val
              sar       this_filter, #1

              ' interpolation of the filtered values
              mov       tmp, this_filter        ' starting point
              mov       tmpt, last_filter
              sub       tmpt, this_filter       ' the delta

              ' scale my value and the delta up to accomodate the fractional bits
              shl       tmp, #12
              shl       tmpt, #12

              or        tmpt, interp_frac       ' into the low 4 bits of tempt
              mov       tmpc, #fraction_bits                ' 4 fractional bits
              ' do the interpolation
:frac_loop    sar       tmpt, #1 wc
        if_c  add       tmp, tmpt
              djnz      tmpc, #:frac_loop

              sar       tmp, #12                ' get the scale correct again

              ' decay the new sample a bit
              mov       sample, tmp
              ' add in some random noise to dither the lowest bit
              mov       tmpt, rnd_seed
              shr       tmpt, sustain_rnd
              add       tmp, tmpt
              ' and perform the scaling of the fraction to be removed
              sar       tmp, sustain_sar wz
              ' and remove it
              sub       sample, tmp

              ' write this sample, and advance
              wrword    sample, ptr_write
              add       ptr_write, #2
              cmpsub    ptr_write, buf_tail wc
        if_c  add       ptr_write, buf_head

              ' loop
              jmp       #loop

adjust_delay
              ' adjust the input period to handle the filter offset
              add       int_delay, #1 << fraction_bits >> 1
              ' and make sure we don't exceed our limit
              max       int_delay, max_delay
              ' calculate the interpolation fraction (next "fraction_bits" bits)
              mov       interp_frac, int_delay
              rev       interp_frac, #32 - fraction_bits
              shr       int_delay, #fraction_bits
              ' and finally recover the delay (address in words...2 bytes per sample)
              ' (leave ptr_write alone, just change ptr_read)
              mov       ptr_read, ptr_write
              add       ptr_read, buf_tail
              sub       ptr_read, buf_head
              sub       ptr_read, int_delay
              sub       ptr_read, int_delay
              ' wrap it as necessary
:wrap         cmpsub    ptr_read, buf_tail wc
        if_c  add       ptr_read, buf_head
        if_c  jmp       #:wrap
              ' done!
adjust_delay_ret
              ret


{=== PASM faux-constants, or initialized variables ===}
prng_add      long      $80808080
middle_val    long      |<15
sample        long      0
pluck_left    long      0
interp_frac   long      0
this_filter   long      0

{=== PASM parameters ===}
ctra_val      long      0
period        long      0
buf_head      long      0
buf_tail      long      0
max_delay     long      0
this_val      long      0
NCO_stride    long      0

{=== PASM scratch variables ===}
sustain_sar   res       1       ' means scale the feedback by (2^N - 1) / (2^N)
rnd_seed      res       1
tmp           res       1
tmpl          res       1
tmpt          res       1
tmpc          res       1
timestamp     res       1
int_delay     res       1
ptr_write     res       1
ptr_read      res       1
pluck_vol     res       1
sustain_rnd   res       1
last_filter   res       1
last_val      res       1
